#!/usr/bin/env python3

import numpy as np
import matplotlib.pyplot as plt

# les vis sont du M3
# tours
vis1 = np.array([0, 1, 2, 3, 2.5, 2, 1.5, 1, 0.5, 0, -0.5, -1, -1.5, -2, -2.5, -3, -3.5])  # +1 tour = devisser, -1 = visser
vis2 = np.array([0, 0.5, 1, 1, 0.5, 0, -0.5, -1, -1.5, -2, -2.5, -3, -3.5, -4])  # +1 tour = devisser, -1 = visser

vis_err = 1/8. # 1/8 de tour

tour2mm = (3.15-1.17) / 9  # mm/tour avec 9 tours mesures au pied a coulisse
tour2um = tour2mm*1e3 # la meme chose en microns / tour

x1s = vis1*tour2um
x2s = vis2*tour2um
x_err = vis_err * tour2um

# courant dans la photodiode
I1 = np.array([56, 33, 11.3, 1.9, 4.7, 12.8, 22.5, 33.5, 49.6, 56.3, 51.5, 35.4, 19.8, 8.4, 3.9, 1.6, 1.1])  # pA
I2 = np.array([56.3, 49.1, 37.6, 39.0, 48.5, 56.3, 55.3, 46.6, 33.3, 20.2, 10.9, 5.0, 1.8, 0.08])  # pA

I_err = 0.2 # pA

fig = plt.figure(0)
plt.scatter(x1s, I1, label="vis #1", marker='o', color='blue')
plt.errorbar(x1s, I1, xerr=x_err, yerr=I_err, fmt='o', color='blue')
plt.scatter(x2s, I2, label="vis #2", marker="o", color='red')
plt.errorbar(x2s, I2, xerr=x_err, yerr=I_err, fmt='o', color='red')
plt.grid()
plt.xlabel("déplacement fibre [$\\mu$m]")
plt.ylabel("$I_{PD}$ [pA]")
plt.axvline(+300., color="k", linestyle="--", label="fiber diameter")
plt.axvline(-300., color="k", linestyle="--")
plt.legend()
plt.show()


## Fit gaussien sur chaque axe

from scipy.optimize import curve_fit

def gaussian(x, A, mu, sigma): #, B):
    return A * np.exp(-(x-mu)**2/(2.*sigma**2)) # + B

# xs_model = np.arange(min(x1s.min(), x2s.min()), max(x1s.max(), x2s.max()), 10) # step 10 microns
xs_model = np.arange(-1000., +1000.+10, 10) # step 10 microns

# Vis #1

A_estimate = I1.max()
# mu_estimate = (x1s * I1).sum() / I1.sum()
mu_estimate = 0.0
var_estimate = ((x1s - mu_estimate)**2 * I1).sum() / I1.sum()
sigma_estimate = var_estimate**.5
# B_estimate = 0.0
print(A_estimate, mu_estimate, sigma_estimate) #, B_estimate)

popt1, pcov1 = curve_fit(gaussian, x1s, I1, p0=[A_estimate, mu_estimate, sigma_estimate]) #, B_estimate])
print("Vis #1:", popt1)

# plt.figure(1)
# plt.scatter(x1s, I1, label="vis #1", marker='+', color='blue')
# plt.plot(xs_model, gaussian(xs_model, *popt1), "b-", label="vis #1 (fit)")
# plt.grid()
# plt.xlabel("déplacement fibre [$\\mu$m]")
# plt.ylabel("$I_{PD}$ [pA]")
# plt.axvline(1e3*0.3, color="k", linestyle="--", label="fiber diameter")
# plt.axvline(1e3*-0.3, color="k", linestyle="--")
# plt.legend()
# plt.show()


# Vis #2

A_estimate = I2.max()
# mu_estimate = (x2s * I2).sum() / I2.sum()
mu_estimate = 0.0
var_estimate = ((x2s - mu_estimate)**2 * I2).sum() / I2.sum()
sigma_estimate = var_estimate**.5
# B_estimate = 0.0
# print(A_estimate, mu_estimate, sigma_estimate)
print(A_estimate, mu_estimate, sigma_estimate) #, B_estimate)

popt2, pcov2 = curve_fit(gaussian, x2s, I2, p0=[A_estimate, mu_estimate, sigma_estimate]) #, B_estimate])
print("Vis #2:", popt2)

# plt.figure(1)
# plt.scatter(x2s, I2, label="vis #2", marker='+', color='red')
# plt.plot(xs_model, gaussian(xs_model, *popt2), "r-", label="vis #2 (fit)")
# plt.grid()
# plt.xlabel("déplacement fibre [$\\mu$m]")
# plt.ylabel("$I_{PD}$ [pA]")
# plt.axvline(1e3*0.3, color="k", linestyle="--", label="fiber diameter")
# plt.axvline(1e3*-0.3, color="k", linestyle="--")
# plt.legend()
# plt.show()



plt.figure(10)
plt.scatter(x1s, I1, label="vis #1", marker='o', color='blue')
plt.errorbar(x1s, I1, xerr=x_err, yerr=I_err, fmt='o', color='blue')
plt.plot(xs_model, gaussian(xs_model, *popt1), "b--", label="vis #1 (fit)")
plt.scatter(x2s, I2, label="vis #2", marker='o', color='red')
plt.errorbar(x2s, I2, xerr=x_err, yerr=I_err, fmt='o', color='red')
plt.plot(xs_model, gaussian(xs_model, *popt2), "r--", label="vis #2 (fit)")
plt.grid()
plt.xlabel("déplacement fibre [$\\mu$m]")
plt.ylabel("$I_{PD}$ [pA]")
plt.axvline(+300., color="k", linestyle="--", label="fiber diameter")
plt.axvline(-300., color="k", linestyle="--")
plt.legend()
plt.show()


