import pandas as pd
import numpy as np
from datetime import datetime
import time
import os
from threading import Thread
import serial

# Create directory to store data
directory = datetime.now().strftime('%Y_%m_%d_%H_%M_%S') + '_full_scan'
# Parent Directory path 
parent_dir = os.getcwd()
# Path 
path = os.path.join(parent_dir, directory) 
os.mkdir(path)

# Acquisition parameters
duration = 20 #sec d'acquisition
interval_keithley6514 = 0.02 #fs = 50 Hz
interval_keysight2987A = 2e-3 #fs = 500 Hz

# ============= CHOPPER =============
# Skr Mini controls chopper wheel
# https://marlinfw.org/meta/gcode/
# chopper = serial.Serial(port='/dev/serial/by-id/usb-STMicroelectronics_MARLIN_STM32G0B1RE_CDC_in_FS_Mode_2066376C5341-if00', baudrate=19200)
# chopper.timeout = 0.1
# chopper.write(b'M92 X48\r')
# chopper.write(b'M203 X160\r')
# chopper.write(b'M201 X200\r')
# chopper_steps = 768 #768 pas = 1 tour
# rotation_frequency = 3
# chopper_rate = chopper_steps*rotation_frequency
# chopper_position = 0
# delta_position = 2500
# chopper.write(b'G92 X0\r') # reset position to zero in software
# chopper.write(f'G0 X{chopper_position} F{chopper_rate}\r'.encode())
# Ask for current position : chopper.write(b'M114\r')

# ============= KEITHLEY =============
from keithley_stardice import *
keithley6514 = Keithley6514('/dev/keithley6514', baudrate=57600, debug=False)
keithley6514.set_timer(interval_keithley6514)
keithley6514.setup(nsamples=int(duration/interval_keithley6514), krange=2e-9, mode='CHAR')
keithley6514.set_output_trigger_line(3)
keithley6514.set_output_trigger()
keithley6514.set_rate(1.0)
print('KEITHLEY CONNECTED')

# ============= KEYSIGHT =============
from keysight import *
keysight2987A = Keysight()
keysight2987A.on()
keysight2987A.sync_tracked_properties()

keysight2987A.set_mode('CHAR')
keysight2987A.set_rang('2e-6')
#keysight2987A.set_mode('CURR')
#keysight2987A.set_rang('2e-6')

keysight2987A.set_nplc(0.1)
keysight2987A.set_nsamples(int(duration/interval_keysight2987A))
keysight2987A.set_delay(0)
keysight2987A.set_interval(interval_keysight2987A)
keysight2987A.set_trigger_out()
print('KEYSIGHT CONNECTED')

# ============= MONOCHROMATOR =============
from zolix import *
M = Zolix()
M.open()
M.set_exit_port(0) # tourner le miroir dans la bonne sortie
grating_id = 2
M.set_grating(grating_id)

# ============= SPECTROGRAPH =============
# from seabreeze.spectrometers import list_devices, Spectrometer
# correct_dark_counts = False
# correct_nonlinearity = False
# spec = Spectrometer.from_first_available()
# spec_wavelengths = spec.wavelengths()
# spec_exp_time = 0.1#0.04
# spec.integration_time_micros(spec_exp_time*1e6) # in 1e-6 seconds
# n_spec = 20

# Threaded class to acquire spectrum in parallel
# precision : seabreeze code is blocking and waits for the exp to complete
class AcquireSpectrum(Thread):
    # constructor
    def __init__(self):
        # execute the base constructor
        Thread.__init__(self)
        # set a default value
        self.spectrum = None
 
    # function executed in a new thread
    def run(self):
        # store data in an instance variable
        self.spectrum = [spec.intensities(correct_dark_counts=correct_dark_counts, correct_nonlinearity=correct_nonlinearity) for i in range(0, n_spec)]
    

# Scan range
scan_wavelengths = np.arange(350, 1070, 10, dtype=int)
# scan_wavelengths = [400, 600, 800]

# Dataframe to contain all read data
df = pd.DataFrame()


# LA GRANDE BOUCLE DEMARRE ICI
# Iterate through each wavelength step
for scan_wavelength in scan_wavelengths:

    #if scan_wavelength > 495 and grating_id==1:
    #    grating_id = 2
    #    M.set_grating(grating_id)
    #    print('GRATING CHANGED')

    # if scan_wavelength > 600 and scan_wavelength < 700:
    #    spec_exp_time = 0.2#0.06
    #    spec.integration_time_micros(spec_exp_time*1e6) # in 1e-6 seconds

    # if scan_wavelength > 700:
    #    spec_exp_time = 0.6#0.08
    #    spec.integration_time_micros(spec_exp_time*1e6) # in 1e-6 seconds

    start_scan = time.time()
        
    M.set_wavelength_and_filter(scan_wavelength)
    time.sleep(0.1)
    print('------------------MONO SET AT {}------------------'.format(scan_wavelength))

    # create a new thread
    # thread_spectrum = AcquireSpectrum()
    # start the thread
    # thread_spectrum.start()

    # chopper_position = chopper_position + delta_position
    # chopper.write(f'G0 X{chopper_position} F{chopper_rate}\r'.encode())
    # print(f'CHOPPER MOVE TO {chopper_position} : ', chopper.readlines())

    # Trigger acquisition on both devices
    keithley6514.set_zch('ON')
    time.sleep(2.)
    keithley6514.set_zch('OFF')
    keysight2987A.reset()

    # da.listen()
    keithley6514.trig_init()
    keysight2987A.write(':INIT:ACQ')

    trig_time = time.time()
    print('TRIG FINISHED : {} s'.format(trig_time-start_scan))

    # timestamps = np.rec.fromrecords(da.get_data_async(), names=['time', 'state'])
    data2 = np.rec.fromrecords(keithley6514.buff_read(), names=['CHAR', 'time', 'flag'])
    read_keithley_time = time.time()
    print('READ KEITHLEY FINISHED : {} s'.format(read_keithley_time-start_scan))
    
    data1 = keysight2987A.read_data()
    read_keysight_time = time.time()
    print('READ KEYSIGHT FINISHED : {} s'.format(read_keysight_time-start_scan))

    # Read and format data
    df_keysight2987A = pd.DataFrame(data1)
    df_keithley6514 = pd.DataFrame(data2)
    #df_timestamps = pd.DataFrame(timestamps)

    # thread_spectrum.join()
    
    # spec_time = time.time()
    # print('SPECTRUM EXPOSURE FINISHED : {} s'.format(spec_time-start_scan))

    d = {'SCAN-WAVELENGTH': scan_wavelength,
         'MC-WAVELENGTH': M.get_wavelength(),
         # 'SPEC-WAVELENGTHS': spec_wavelengths,
         #'SPEC-FLUX': np.array(thread_spectrum.spectrum),
         #'SPEC-EXP-TIME': spec_exp_time,
         #'correct_dark_counts': correct_dark_counts,
         #'correct_nonlinearity': correct_nonlinearity,
         #'SC-CURR': np.array(df_keysight2987A['CURR']),
         'SC-CHAR': np.array(df_keysight2987A['CHAR']),
         'SC-TIME': np.array(df_keysight2987A['time']),
         'PD-CHAR': np.array(df_keithley6514['CHAR']),
         'PD-TIME': np.array(df_keithley6514['time']),
	 'GRATING-ID': grating_id}
         #'ROTATION-SPEED': rotation_frequency*2,}

    df = df.append(d, ignore_index=True)
    print(df.tail(10))
    
    end_scan = time.time()
    print('WAVELENGTH SCAN FINISHED : {}'.format(end_scan-start_scan))
    
# Save to different file formats
df.to_csv(path+'/'+datetime.now().strftime('%Y_%m_%d_%H_%M_%S')+'_full_scan.csv', index=False)
df_rec = df.to_records()
np.save(path+'/'+datetime.now().strftime('%Y_%m_%d_%H_%M_%S')+'_full_scan.npy', df_rec)
