import io
import numpy as np
import serial
import string
import sys
import time
import types

k6514_error_map = {
    '-440': 'Query unterminated after indefinite response EE ',
    '-430': 'Query deadlocked                             EE ',
    '-420': 'Query unterminated                           EE ',
    '-410': 'Query interrupted                            EE ',
    '-363': 'Input buffer overrun                         EE ',
    '-362': 'Framing error in program message             EE ',
    '-361': 'Parity error in program message              EE ',
    '-360': 'Communications error                         SYS',
    '-350': 'Queue overflow                               EE ',
    '-330': 'Self-test failed                             EE ',
    '-314': 'Save/recall memory lost                      EE ',
    '-315': 'Configuration memory lost                    EE ',
    '-285': 'Program syntax error                         EE ',
    '-284': 'Program currently running                    EE ',
    '-282': 'Illegal program name                         EE ',
    '-281': 'Cannot create program                        EE ',
    '-260': 'Expression error                             EE ',
    '-241': 'Hardware missing                             EE ',
    '-230': 'Data corrupt or stale                        EE ',
    '-225': 'Out of memory                                EE ',
    '-224': 'Illegal parameter value                      EE ',
    '-223': 'Too much data                                EE ',
    '-222': 'Parameter data out of range                  EE ',
    '-221': 'Settings conflict                            EE ',
    '-220': 'Parameter error                              EE ',
    '-215': 'Arm deadlock                                 EE ',
    '-214': 'Trigger deadlock                             EE ',
    '-213': 'Init ignored                                 EE ',
    '-212': 'Arm ignored                                  EE ',
    '-211': 'Trigger ignored                              EE ',
    '-210': 'Trigger error                                EE ',
    '-202': 'Settings lost due to rtl                     EE ',
    '-201': 'Invalid while in local                       EE ',
    '-200': 'Execution error                              EE ',
    '-178': 'Expression data not allowed                  EE ',
    '-171': 'Invalid expression                           EE ',
    '-170': 'Expression error                             EE ',
    '-168': 'Block data not allowed                       EE ',
    '-161': 'Invalid block data                           EE ',
    '-160': 'Block data error                             EE ',
    '-158': 'String data not allowed                      EE ',
    '-154': 'String too long                              EE ',
    '-151': 'Invalid string data                          EE ',
    '-150': 'String data error                            EE ',
    '-148': 'Character data not allowed                   EE ',
    '-144': 'Character data too long                      EE ',
    '-141': 'Invalid character data                       EE ',
    '-140': 'Character data error                         EE ',
    '-128': 'Numeric data not allowed                     EE ',
    '-124': 'Too many digits                              EE ',
    '-123': 'Exponent too large                           EE ',
    '-121': 'Invalid character in number                  EE ',
    '-120': 'Numeric data error                           EE ',
    '-114': 'Header suffix out of range                   EE ',
    '-113': 'Undefined header                             EE ',
    '-112': 'Program mnemonic too long                    EE ',
    '-111': 'Header separator error                       EE ',
    '-110': 'Command header error                         EE ',
    '-109': 'Missing parameter                            EE ',
    '-108': 'Parameter not allowed                        EE ',
    '-105': 'GET not allowed                              EE ',
    '-104': 'Data type error                              EE ',
    '-103': 'Invalid separator                            EE ',
    '-102': 'Syntax error                                 EE ',
    '-101': 'Invalid character                            EE ',
    '-100': 'Command error                                EE ',
    '+000': 'No error                                     SE ',
    '+101': 'Low limit 1 failed                           SE ',
    '+102': 'High limit 1 failed                          SE ',
    '+103': 'Low limit 2 failed                           SE ',
    '+104': 'High limit 2 failed                          SE ',
    '+105': 'Active limit tests passed                    SE ',
    '+106': 'Reading available                            SE ',
    '+107': 'Reading overflow                             SE ',
    '+108': 'Buffer available                             SE ',
    '+109': 'Buffer full                                  SE ',
    '+200': 'Operation complete                           SE ',
    '+300': 'Device calibrating                           SE ',
    '+303': 'Device sweeping                              SE ',
    '+305': 'Waiting in trigger layer                     SE ',
    '+306': 'Waiting in arm layer                         SE ',
    '+310': 'Re-entering the idle layer                   SE ',
    '+408': 'Questionable calibration                     SE ',
    '+414': 'Command warning                              SE ',     
    '+500': 'Date of calibration not set                  EE ',
    '+501': 'Next date of calibration not set             EE ',
    '+502': 'Calibration data invalid                     EE ',
    '+507': 'Measurement offset data invalid              EE ',
    '+508': 'Measurement gain data invalid                EE ',
    '+509': 'Not permitted with cal locked                EE ',
    '+510': 'Not permitted with cal un-locked             EE ',
    '+511': 'Voltage offset not converging                EE ',
    '+512': 'Current offset not converging                EE ',
    '+602': 'GPIB address lost                            EE ',
    '+603': 'Power-on state lost                          EE ',
    '+604': 'DC calibration data lost                     EE ',
    '+605': 'Calibration dates lost                       EE ',
    '+606': 'GPIB communication language lost             EE ',
    '+700': 'Invalid system communication                 EE ',
    '+701': 'ASCII only with RS-232                       EE ',
    '+800': 'Illegal with storage active                  EE ',
    '+801': 'Insufficient vector data                     EE ',
    '+804': 'Expression list full                         EE ',
    '+805': 'Undefined expression exists                  EE ',
    '+806': 'Expression not found                         EE ',
    '+807': 'Definition not allowed                       EE ',
    '+808': 'Expression cannot be deleted                 EE ',
    '+811': 'Not an operator or number                    EE ',
    '+812': 'Mismatched parentheses                       EE ',
    '+813': 'Not a number of data handle                  EE ',
    '+814': 'Mismatched brackets                          EE ',
    '+815': 'Too many parentheses                         EE ',
    '+816': 'Entire expression not parsed                 EE ',
    '+817': 'Unknown token                                EE ',
    '+818': 'Error parsing mantissa                       EE ',
    '+819': 'Error parsing exponent                       EE ',
    '+820': 'Error parsing value                          EE ',
    '+821': 'Invalid data handle index                    EE ',
    '+830': 'Invalid with INFinite ARM:COUNT              EE ',
    '+831': 'Invalid with INFinite TRIG:COUNT             EE ',
    '+900': 'Internal system error                        EE ',
    '+950': 'Rdg overflow                                 SE ',
    '+951': 'Rdg ready                                    SE ',
    '+952': 'Buffer full                                  SE ',
    '+953': 'IDDC error                                   EE ',
    '+954': 'IDDCO error                                  EE ',
    '+955': 'Trig overrun                                 EE ',
    '+956': 'No remote                                    EE ',
    '+957': 'Number error                                 EE ',
    '+958': 'DDC ready                                    SE ',
    '+960': 'DDC Mode IDDC Error                          EE ',
    '+961': 'DDC Mode IDDCO Error                         EE ',
    '+962': 'DDC Ready                                    SE ',
    '+963': 'DDC Reading Done                             SE ',
    '+965': 'DDC Buffer Full                              SE ',
    '+966': 'DDC Reading overflow                         SE ',}

    
#    def check_error_status(self):
#        """
#        Get the various error status, clear them, and may
#        raise the corresponding Exception (if needed).
#        """
#
#        self.reopen_if_needed()
#        self.purge()
#
#        # First, get (and clear) the Standard Event Status Register
#
#        time.sleep(1)
#
#        command = "*ESR?"
#        self.write(command)
#        answer = self.read()
#        if not(answer):
#            raise IOError("Keithley6514: *ESR? command failed (no answer).")
#        try:
#            esr = int(answer)
#        except ValueError:
#            raise IOError("Keithley6514: " +
#                          "*ESR? command failed (invalid answer [1]).")
#
#
#        self.clear()
#
#        if (esr & 0x20):
#            # Command Error. Set when a syntax type error is detected
#            # in a command from the bus. The parser is reset and parsing
#            # continues at the next byte in the input stream.
#            raise MultimeterError("Keithley6514: PowerSupply Command Error. " +
#                          "A syntax type error has been detected " +
#                          "in a command from the bus. " +
#                          "The parser has been reset and parsing " +
#                          "continues at the next byte in the input stream.")
#
#        if (esr & 0x10):
#            # Execution Error. Set when an error is encountered while
#            # attempting to execute a completely parsed command.
#            # The appropriate error number will be reported in 
#            # the Execution Error Register
#
#            # Now reading (and clearing) the Execution Error Register
#
#            command = "EER?"
#            self.write(command)
#            answer = self.read()
#            if not(answer):
#                raise IOError("Keithley6514: EER? command failed (no answer).")
#            try:
#                eer = int(answer)
#            except ValueError:
#                raise IOError("Keithley6514: " +
#                              "EER? command failed (invalid answer [1]).")
#
#            if (eer >= 1) and (eer <= 99):
#                raise MultimeterError("Keithley6514: Hardware error.")
#
#            if eer == 116:
#                raise MultimeterError("Keithley6514: Invalid recall of data.")
#
#            if eer == 117:
#                raise MultimeterError("Keithley6514: Corrupted internal data.")
#
#            if eer == 120:
#                raise MultimeterError("Keithley6514: " +
#                              "Numerical specified value was too big " + 
#                              "or too small.")
#
#            if eer == 123:
#                raise MultimeterError("Keithley6514: " +
#                              "Illegal recall requested.")
#
#            if eer == 124:
#                raise MultimeterError("Keithley6514: " +
#                              "Illegal range change requested.")
#
#
#        if (esr & 0x08):
#            # Verify Timeout Error. Set when a parameter is set with 'verify'
#            # specified and the value is not reached within 5 secs.
#            raise MultimeterError("Keithley6514: Verify Timeout Error. " +
#                          "Set when a parameter is set with 'verify' " +
#                          "specified and the value is not reached " +
#                          "within 5 secs.")
#
#        if (esr & 0x04):
#            # Query Error. Appropriate number is reported in 
#            # the Query Error register
#            raise MultimeterError("Keithley6514: Query Error.")



def auto_method(self, command):
    f = string.Formatter()
    g = f.parse(command)
    fields = [entry[1] for entry in g if entry[1] not in ['', 'mode']]
    def meth(self, keys={}):
        if isinstance(keys, dict):
            self.settings.update(keys)
        elif len(fields) == 1:
            self.settings[fields[0]] =keys
        else:
            raise ValueError('Command needs the following values: %s' % str(fields))
        self.write(command.format(**self.settings))
    meth.__doc__= command
    return types.MethodType(meth, self)

def auto_getter(self, name):
    def meth(self):
        return self.settings[name]
    return types.MethodType(meth, self)

class Keithley6514(object):
    error_states = {0x20:'Syntax error'}
    commands = {'reset':"*RST",
                'set_zch': "SYST:ZCH {zch}",
                'set_autozero': "SYST:AZER {autozero}",
                'set_nsamples': 'ARM:COUN {nsamples:d}',
                'set_rate': ':SENS:{mode}:NPLC {nplc:f}',
                'set_range': '{mode}:RANG {range}',
                'set_func': "FUNC '{mode}'", 
                'disp_off': ':DISP:ENAB OFF',
                'set_timer': ":ARM:TIM {timer}",
                'set_freq': ":SYST:LFR 50",
                'reset_timestamp': ":SYST:TIME:RES",
                'set_output_trigger_line': 'TRIG:TCON:ASYN:OLIN {output_trigger_line:d}',
                'set_output_trigger': 'TRIG:TCON:ASYN:OUTP SENS',
    }

    def __init__(self, dev, baudrate=57600, EOL='\r', debug=True, buff=2500):
        self.debug = debug
        self.serial_port = serial.Serial(dev, baudrate=baudrate)
        self.EOL = EOL
        self.io = io.TextIOWrapper(self.serial_port, write_through=True, line_buffering=True, newline=self.EOL)
        self.io._CHUNK_SIZE = 1
        self.buff=buff
        self.write("*IDN?")
        self.id = self.read()
        self.settings = {'mode':'CHAR',
                         'nplc':0.1,
                         'range':'2e-6',
                         'median_filter':'OFF',
                         'digital_filter':'OFF',
                         'nsamples':1,
                         'timer': 0.05,
                         'zch': 'OFF',
                         'autozero': 0,
                         'output_trigger_line': 3,
        }
        for name in self.commands:
            setattr(self, name, auto_method(self, self.commands[name]))
        
        for name in self.settings:
            setattr(self, 'get_'+name, auto_getter(self, name))

    def write(self, command):
        """
        Send a command through the serial port.
        """
        if self.debug: print("Keithley6514: Sending command [" + command + "]", file=sys.stderr)
        #self.serial_port.write((command + self.EOL).encode('ASCII'))
        self.io.write((command + self.EOL))


    def read(self, timeout = None):
        """
        Read the answer from the serial port.
        Return it as a string.

        If <timeout> is specified, the function will wait
        for data with the specified timeout (instead of the default one). 
        """
        
        if self.debug: print("Keithley6514: " + \
                "Reading serial port buffer", file=sys.stderr)

        #answer = self.serial_port.readline() # return buffer
        answer = self.io.readline() # return buffer
        
        # remove end of line
        #answer = answer.strip()
        if self.debug:
            print("Keithley6514: " + \
                  "Received [" + answer + "]", file=sys.stderr)

        return answer

    
        
    def get_settings(self):
        return self.settings
    
    def setup(self, nsamples=1, krange=2.E-7, mode='CHAR'):
        self.reset()
        time.sleep(1)
        self.settings['mode']=mode
        self.set_zch('ON')
        self.set_func()
        self.set_range({'range':krange})
        self.set_nsamples({'nsamples':nsamples})
        self.set_rate()
        self.write('ARM:SOUR TIM') 
        self.set_timer()
        self.disp_off()
        self.set_zch('OFF')
        self.set_autozero(0)

    def trig_init(self):
        self.write('TRAC:FEED:CONT NEV')
        self.write(f'TRAC:POIN {self.buff:d}')
        self.write('TRAC:CLE')
        self.write('TRAC:FEED:CONT NEXT')
        self.write('INIT')
        
    def buff_read(self):
        self.write("TRAC:DATA?")
        data_str = self.read()
        return np.array(data_str.split(',')).astype(float).reshape([-1, 3])
    
if __name__ == '__main__':
    k = Keithley6514('/dev/keithley6514', baudrate=57600)
    k.setup(nsamples=100)
    k.trig_init()
    data = k.buff_read()
    
