import telnetlib
import time
import numpy as np
import struct
import array

class Config():
    pass

class Keysight():
    tracked_properties = ['mode',
                          'rang',
                          'nplc',
                          'nsamples',
                          'interval',
                          'delay',]
                          #'aper']
    
    expected_params = {'mode':'CURR',
                       'rang':'2e-6' ,
                       'nplc':0.1,
                       'nsamples':100,
                       'delay':0,
                       'interval':2e-3,}
                       #'aper':'AUTO'}

    def __init__(self, keysight_addr='169.254.5.2', keysight_port=5025):
        self.config = Config
        self.config.keysight_addr = keysight_addr
        self.config.keysight_port = keysight_port
        
    def _config(self):
        pass
        #self.client = telnetlib.Telnet(self.config.keysight_addr, self.config.keysight_port)
        
    def sync_tracked_properties(self):
        print('Setting parameters')
        for p in self.tracked_properties:
            setattr(self, f'_{p}', self.expected_params[p])
        for p in self.tracked_properties:
            getattr(self, f'set_{p}')(self.expected_params[p])
        
    def write(self, message):
        self.client = telnetlib.Telnet(self.config.keysight_addr, self.config.keysight_port)
        try:
            self.client.write((message+'\r\n').encode())
        except EOFError:
            self._config()
            self.client.write((message+'\r\n').encode())
        self.client.close()
        time.sleep(0.1)
        
    def read(self, message):
        self.client = telnetlib.Telnet(self.config.keysight_addr, self.config.keysight_port)
        self.client.write((message+'\r\n').encode())
        try:
            return self.client.read_until(b'\n', timeout=10)
        except EOFError:
            self._config()
            return self.client.read_until(b'\n', timeout=10)
        self.client.close()
        time.sleep(0.1)
        
    def on(self):
        self.write(':INP ON')

    def off(self):
        self.write(':INP OFF')

    def get_power(self) :
        return self.read(':INP?').decode()[0]
    
    def acquire(self):
        t_acq = self._nsamples*self._interval
       # print(f'nsample = {self._nsamples}, nplc = {self._nplc}')#, f_plc = {f_plc}')
        self.write(':INIT:ACQ')
        
        for t in range(int(t_acq*100) + 1) :
            print(f'Acquisition {t/100:.2f}/{t_acq:.2f} s', end = '\r')
            time.sleep(1e-2)

        print('\nAcquisition finished')
            
    def read_binary_data(self):
        """
        d = self.read(f':FORM:DATA REAL,32')
        n = struct.unpack('>I', d[:4])
        data = struct.unpack(f'>{n}f', d[4:-1])
        self.write(f':FORM:DATA ASC')
        return data
        """
        print('Reading the data')
        self.write(f':FORM:DATA REAL,32')
        # Read time
        t = self.read(f':FETC:ARR:TIME?')
        t = array.array('f', t[2:])
        t.byteswap()
        t = np.array(t.tolist())
        # Read data
        d = self.read(f':FETC:ARR:{self._mode}?')
        d = array.array('f', d[6:])
        d.byteswap()
        d = np.array(d.tolist())
        
    def read_data(self):
        print('Reading the data')
        self.write(f':FORM:DATA ASC')
        t = self.read(f':FETC:ARR:TIME?')
        t = np.array(t.split(b',')).astype(float)
        d = self.read(f':FETC:ARR:{self._mode}?')
        d = np.array(d.split(b',')).astype(float)
        return np.rec.fromarrays([t,d], names=['time', self._mode])
         
    def reset(self):
        self.write('SENS:CHAR:DISCharge')

    def set_trigger_out(self) :
        self.write(f'TRIG:ACQ:TOUT 1')
        self.write(f'TRIG:ACQ:TOUT:SIGN TOUT')        

    def set_mode(self, mode):

        if mode not in ['CURR', 'CHAR', 'VOLT', 'RES'] :
            raise ValueError("Mode should be one of these valors : 'CURR', 'CHAR', 'VOLT', 'RES'")

        mode = str(mode)
        self.write(f'SENSe:FUNCtion:ON "{mode}"')
        self._mode=mode
        self.set_nplc(self._nplc)
        self.set_rang(self._rang)
        self.set_nsamples(self._nsamples)

    def get_mode(self):
        return self.read(f'SENS:FUNC?')

    def set_aper(self, aper):
        if aper == 'AUTO':
            self.write(f'{self._mode}:APER:AUTO ON')
        else:
            self.write(f'{self._mode}:APER:AUTO OFF')
            self.write(f'{self._mode}:APER {aper}')
        self._aper = aper

    def get_aper(self):
        return self.read(f'SENS:{self._mode}:APER?')    
        
    def set_nplc(self, nplc):
        if nplc == 'AUTO' :
            self.write(f':SENS:{self._mode}:NPLC:AUTO ON')
        else:
            self.write(f':SENS:{self._mode}:NPLC:AUTO OFF')
            self.write(f':SENS:{self._mode}:NPLC {nplc}')
        self._nplc = nplc

    def get_nplc(self):
        return self.read(f'SENS:{self._mode}:NPLC?')

    def set_rang(self, charge_range):
        '''value for current measurement : 2e-12 A to 20e-3 A
        value for charge measurement : 2e-9 C to 2e-6 C'''
        if charge_range == 'AUTO' :
            self.write(f'SENS:{self._mode}:RANG:AUTO ON')
        else :
            charge_range = float(charge_range)
            self.write(f'SENS:{self._mode}:RANG:AUTO OFF')
            self.write(f'SENS:{self._mode}:RANG {charge_range}')            
        self._rang = charge_range

    def get_rang(self) :
        return self.read(f'SENS:{self._mode}:RANG?')
            
    def set_delay(self, delay_time) :
        delay_time = float(delay_time)
        self.write(f':TRIG:ACQ:DEL {str(delay_time)}')
        self.write(f':TRIG:SOUR TIM')
        self._delay = delay_time

    def get_delay(self):
        return self.read(f'TRIG:ACQ:DEL?')

    def set_interval(self, interval) :
        interval = float(interval)
        self.write(f':TRIG:ACQ:TIM {str(interval)}')
        self._interval = interval
        
    def get_interval(self) :
        return self.read(':TRIG:ACQ:TIM?')
        
    def set_nsamples(self, nsamples=5500):
        nsamples = int(nsamples)
        self.write(f':TRIG:ACQ:COUN {nsamples}')
        self._nsamples = nsamples

    def get_nsamples(self):
        return self.read(f':TRIG:ACQ:COUN?').decode()

    def before_exposure(self, exposure, fitsbuilder):
        Subsystem.before_exposure(self, exposure, fitsbuilder)
        res = self._process_params(exposure)
        self.set_trigger_out()
        self.on()
        
    def exposure(self, exposure, fitsbuilder):
        self.reset()
        self.acquire()
        
    def after_exposure(self, exposure, fitsbuilder):
        super().after_exposure(exposure, fitsbuilder)
        res = self.read_data()
        fitsbuilder.add_table(self.name, res)


if __name__=='__main__':
    
    k = Keysight()
    duration = 20
    interval = 2e-3
    k.on()
    k.sync_tracked_properties()
    k.set_mode('CHAR')#CURR
    k.set_rang('2e-6')#2e-6
    k.set_nplc(0.1)
    k.set_nsamples(int(duration/interval))
    k.set_delay(0)
    k.set_interval(interval)
    k.reset()
    k.acquire()
    res = k.read_data()
    
